
************************************************************************
*
*	6 8 0 0 0  P r o c e s s o r  D e f i n i t i o n s
*	===================================================
*
*	These are the definitions for the 68000 and the ATARI ST.
*		( Version Nov. 6 / 1992 )
************************************************************************
		.text
enable:		.equ	$2500		* Enable interrupts value for SR
disable:	.equ	$2700		* Disable interrupts SR value
*
*		Vector definitions
*
trapvec:	.equ	$80		* Trap 0 vector address
*
*		Hardware Locations
*
giselect:	.equ	$ffff8800	* PSG addresses
giread:		.equ	$ffff8800	*
giwrite:	.equ	$ffff8802	*
*
MK68901:	.equ	$fffffa01	* MFP addresses
IERB:		.equ	$fffffa09	*
IMRB:		.equ	$fffffa15	*
VR:		.equ	$fffffa17	*
TCDCR:		.equ	$fffffa1d	*
TCDR:		.equ	$fffffa23	*
*
		.page

************************************************************************
*
*		L B O S  E r r o r  C o d e s
*		=============================
*
* The following error codes can be returned in D0.L if an error occurs
* during the processing of an SVC.
*
************************************************************************

e_badsvc:	.equ	-1		* Illegal SVC code in D0
e_nomcb:	.equ	-2		* No MCB available for request
e_badtask:	.equ	-3		* Illegal task number
e_badmcb:	.equ	-4		* Illegal MCB address on reply

************************************************************************
*
*		S V C  C o d e  E q u a t e s
*		=============================
*
* These symbols are used by O/S tasks for referencing SVC codes.
*
************************************************************************

SVCNEW:		.equ	0		* New task SVC
SVCPRT:		.equ	1		* Print on terminal SVC
SVCDLY:		.equ	2		* Delay SVC
SVCSND:		.equ	3		* Send SVC
SVCRCV:		.equ	4		* Receive Svc
SVCRPLY:	.equ	5		* Reply SVC
		.page




************************************************************************
*
*		T C B  (Task Control Block) Layout
*		==================================
*
* Each task in the LBOS system is represented by a data structure called
* a "Task Control Block".  This area is used by the operating system to
* store task-specific data.
*
* Offset			Contents			Name
* ======			========			====
*
*	  +---------------------------------------------+
*    0	  | Pointer to next TCB or Zero if no next TCB	|	t_link
*	  +---------------------------------------------+
*    4	  | Pointer to messages waiting to be RECEIVEd	|	t_msg
*	  +---------------------------------------------+
*    8	  | Pointer to messages awaiting REPLY		|	t_rply
*	  +---------------------------------------------+
*   12	  | Contents of task's USP when task is waiting	|	t_usp
*	  +---------------------------------------------+
*   16	  | Contents of task's SSP when task is waiting	|	t_ssp
*	  +---------------------------------------------+
*   20	  | Task flags word				|	t_flag
*	  +---------------------------------------------+
*   24	  | Stack area for system			|
*	  +---------------------------------------------+
*
* The task's registers, PC, and status register are stored on the system
* stack area within the TCB when the task is not running.
*
************************************************************************
*
*		Assembler definitions:
*
stacksize:	.equ	$100		* Size of system stack
*
*		TCB offsets
*
t_link:		.equ	0		* Link word for task list thread
t_msg:		.equ	t_link+4	* Link to waiting messages
t_rply:		.equ	t_msg+4		* Link to messages received
t_usp:		.equ	t_rply+4	* User Stack Pointer area
t_ssp:		.equ	t_usp+4		* System Stack Pointer area
t_flag:		.equ	t_ssp+4			* Flag word
t_stack:	.equ	t_flag+4+stacksize	* Stack area
tcbsize:	.equ	t_stack			* Size of TCB in bytes
*
*		Flag word bits:
*
IWAIT:		.equ	0		* Waiting on interrupt
MWAIT:		.equ	1		* Waiting on message completion
RWAIT:		.equ	2		* Waiting on message receipt
		.page












************************************************************************
*
*		M e s s a g e  C o n t r o l  B l o c k
*               =======================================
*
* LBOS performs system services by routing a message from the requesting
* task to a task which performs the appropriate function.   Each message
* is represented by a fixed length data area called a "Message Control
* Block". This area has the following layout:
*
* Offset			Contents			Name
* ======			========			====
*
*	  +---------------------------------------------+
*     0	  | Link to next MCB or Zero if no next MCB	|	m_link
*	  +---------------------------------------------+
*     4	  | Buffer address (if required)		|	m_buff
*	  +---------------------------------------------+
*     8	  | 32-bit count (byte count or delay time)	|	m_count
*	  +---------------------------------------------+
*    12	  | Requesting TCB address			|	m_rtcb
*	  +---------------------------------------------+
*    16	  | Returned Status				|	m_stat
*	  +---------------------------------------------+
*
* A message is sent using the SEND SVC and received using the RECEIVE
* SVC.    The issuing task is suspended until the serving task issues
* a REPLY SVC.
*
************************************************************************
*
*		Assembler Definitions:
*
m_link:		.equ	0		* Link word
m_buff:		.equ	m_link+4	* Buffer address for I/O
m_count:	.equ	m_buff+4	* 32-bit count
m_rtcb:		.equ	m_count+4	* Requesting TCB addresss
m_stat:		.equ	m_rtcb+4	* Status word
mcbsize:	.equ	m_stat+4	* Size of MCB
		.page


























************************************************************************
*
*		L B O S  I n i t i a l i z a t i o n  C o d e
*		=============================================
*
* Routine "init" performs initialization prior to LBOS receiving
* control. Actions performed are:
*
*	1).	Initialize vector area to appropriate interrupt routines
*	2).	Initialize MFP & PSG
*	3).	Initialize TCB area
*	4).	Initialize MCB free list
*	5).	Jump to dispatcher
*
************************************************************************

init:
	move.w	#-1,d0			*   Delay a few seconds to
clean:	move.w	#50,d1			*   allow the TOS to clean
house:	nop				*   up the last disk access.
	dbra	d1,house		*
	dbra	d0,clean		*
	clr.l	-(sp)			*   No exit from LBOS
	move.w	#$20,-(sp)		*	Set Supervisor
	trap	#1			*		Mode
	move.w	#disable,sr		*   Disallow interrupts
	move.l	#trap0,trapvec		*** Set trap vector
	move.l	#timeint,$114		*** Set Timer C vector
	move.l	#termint,$100		*** Set Centronics Busy vector
	moveq	#$0,d0			*** Initialize MFP
	movea.l	#MK68901,a0		***	"
	movep.l	d0,6(a0)		***	"
	movep.l	d0,14(a0)		***	"
	movep.l	d0,22(a0)		***	"
	movep.l	d0,30(a0)		***	"
	movep.w	d0,2(a0)		***	"
	move.b	#$21,IMRB		*** Timer C & Busy mask enabled
	move.b	#$40,VR			*** Mfp base=$100, AEI mode
	move.b	#$60,TCDCR		*** Timer C prescaler=100
	move.b	#207,TCDR		*** Timer C data
	move.b	#$07,giselect		*** Select register 7 in PSG
	move.b	#$c0,giwrite		*** Port A & B DDR = outputs
	move.b	#$0e,giselect		*** Select Port A
	move.b	#$37,giwrite		*** Initialize port A
	move.b	#$0f,giselect		*** Select Port B for output
	lea	initdata,a0		*** a0 -> TCB init area
	lea	tcblist,a3		*** a3 -> "previous TCB"
inittcb:
	move.l	(a0)+,a1		*** a1 -> uninitialized tcb area
	cmpa.l	#0,a1			*** Done yet?
	beq	tcbdone			*** EQ => done with tcb area
	clr.l	t_usp(a1)		*** Clear task's USP
	lea.l	t_stack(a1),a2		*** a2 -> End of stack area
	move.l	(a0)+,-(a2)		*** Push task's "PC"
	tst.w	(a0)+			*** Skip high word of SR
	move.w	(a0)+,-(a2)		*** Push task's SR
	movem.l	d0-d7/a0-a6,-(a2)	*** save "registers"
	move.l	a2,t_ssp(a1)		*** Set kernel stack
	move.l	a1,t_link(a3)		*** Set previous tcb link word
	move.l	a1,a3			*** Save previous tcb address
	bra	inittcb			*** Loop till done
tcbdone:
	clr.l	t_link(a3)		*** Null out last word



*
*		Init Message control block area
*
	move.l	#nmcbs-1,d0		*** d0 = number of mcbs to init
	lea	mcbarea,a0		*** a0 -> area
	move.l	a0,mcblist		*** Init list head
mcbloop:
	move.l	a0,a1			*** Save current address
	lea	mcbsize(a0),a0		*** a0 -> Next mcb
	move.l	a0,m_link(a1)		*** Set link word
	dbra	d0,mcbloop		*** repeat till done
	clr.l	m_link(a1)		*** set last word to 0
*
*		Prepare to enter dispatcher
*
	clr.l	current			*** No current task
	jmp	dispatch		*** Go to dispatcher

















































************************************************************************
*
*		L B O S  C a l l i n g  C o n v e n t i o n s
*		=============================================
*
* LBOS supports 6 SVC's (SuperVisor Calls). An SVC is requested by 
* loading the appropriate function code into D0.L  and executing a
* TRAP #0 instruction.  The following functions are available:
*
* Function		D0	      A0	    D1		   D2
* =====================	==	==============	===========	========
* Newtask		 0	      N/A	    N/A		  N/A
* Printline on terminal	 1	Buffer Address	Byte Count	  N/A
* Delay			 2	      N/A	Delay Count	  N/A
* Send Message		 3	Buffer Address	Count		 Task #
* Receive Message	 4	      N/A	    N/A		  N/A
* Reply to Message	 5	MCB Address	    N/A		  N/A
*
* The entry point for SVC routines is "trap0:".
*
************************************************************************

trap0:
	movem.l	d0-d7/a0-a6,-(a7)	*   Save all registers
	cmp.l	#maxsvc,d0		*   Is d0 in range?
	bhis	badsvc			*   If HIS, no, return error
	lsl.l	#2,d0			*   multiply d0 by 4
	lea.l	svctab,a6		*   a6 -> beginning of table
	move.l	0(a6,d0.l),a6		*   a6 -> svc routine
	jmp	(a6)			*   enter svc routine
*
*	If function number is out of range, return -1.
*
badsvc:	movem.l	(sp)+,d0-d7/a0-a6	*   Restore registers
	move.l	#e_badsvc,d0		*   load return code
	rte				*   Return to task code
*
*	Table of SVC routine addresses, in function code order.
*
	.data
svctab:	.dc.l	newtask			* 0 Enter dispatcher
	.dc.l	prtline			* 1 Terminal output
	.dc.l	delay			* 2 Real-time delay
	.dc.l	send			* 3 Send a message
	.dc.l	recv			* 4 Receive a message
	.dc.l	reply			* 5 Reply to a message
maxsvc:	.equ	(*-svctab)/4		*   Max SVC number + 1
	.text
	.page

















************************************************************************
*
*		N e w t a s k  S V C  R o u t i n e
*		===================================
*
* SVC "newtask" is used to suspend the execution of the present task and
* schedule execution of the next task. This routine is called by other
* SVC routines when the current task becomes non-dispatchable.
*
* Entry Conditions:
*
*	Registers and PC/SR stored on stack in standard order.
*
* Exit Conditions:
*
*	Task Content saved.  Exit to dispatcher.
*
************************************************************************

newtask:
	move.l	current,a0		*  a0 -> current TCB
	cmpa.l	#0,a0			*  check to be sure there is one
	beq	dispatch		*  EQ => No current task
	move.l	usp,a1			*  Fetch user stack pointer
	move.l	a1,t_usp(a0)		*  Save in TCB
	move.l	a7,t_ssp(a0)		*  Save supervisor stack pointer
	clr.l	current 		*  Indicate no active task
*
* Here we try to dispatch the highest priority task.  If none, wait
* until there is one.
*
dispatch:
	move.l	#tcblist,a0		*   a0 -> list header
	move.w	#disable,sr		*   disable interrupts
d_loop:	move.l	t_link(a0),a0		*** a0 -> Next TCB
	cmpa.l	#0,a0			*** Is there a next tcb?
	beq	none			*** EQ => tasks not ready. Wait.
	tst.l	t_flag(a0)		*** Can this task run?
	bne	d_loop			*** If NE, no, try next one
	move.l	a0,current		*** Set current task
	move.l	t_usp(a0),a1		*** a1 = user stack pointer
	move.l	a1,usp			*** set usp
	move.l	t_ssp(a0),a7		*** set kernel stack pointer
dorte:	movem.l	(a7)+,d0-d7/a0-a6	*** restore task's registers
	rte				*** (Re) enter task code
*
* Here if no ready tasks.  Enable interrupts and wait.
*
none:	stop	#$2500			*   Wait for interrupt
	bra	dispatch		*   Try to dispatch again
	.page















************************************************************************
*
*			L B O S  D a t a  A r e a
*			=========================
*
* This area contains the data used by the operating system.
*
************************************************************************

	.data				*   Initialized data
*
*	Data for initialization routine
*
	.globl	app1epa,app2epa		*   Applications entry points

initdata:				*   Initialization Data for:
	.dc.l	timetask,timeepa,$2500	*   Timer task tcb
	.dc.l	termtask,termepa,$2500	*   Terminal task tcb
	.dc.l	app1task,app1epa,$0500	*   Application task #1
	.dc.l	app2task,app2epa,$0500	*   Application task #2
	.dc.l	0			*   End of list marker

		.bss			*   Uninitialized data

current:	.ds.l	1		*   -> Current task's TCB
tcblist:	.ds.l	1		*   -> TCBs linked by priority
*
*		TCB storage area
*
		.even
timetask:	.ds.b	tcbsize		*   Timer task TCB area
termtask:	.ds.b	tcbsize		*   Terminal task TCB area
app1task:	.ds.b	tcbsize		*   Application task #1
app2task:	.ds.b	tcbsize		*   Application task #2
*
*		MCB storage area
*
mcblist:	.ds.l	1		*   -> Free list of MCBs
nmcbs:		.equ	5		*   Number of free mcbs
		.even
mcbarea:	.ds.b	mcbsize*nmcbs	*   Area for MCBs
		.text
		.page























************************************************************************
*
*		S e n d  M e s s a g e  S V C
*		=============================
*
* Routine "send" is entered via TRAP or a JUMP from another SVC handler.
* This routine allocates a message control block,  and marks the task as
* non-dispatchable. It clears the RWAIT (Receive Wait) bit in the target
* task.
*
************************************************************************

send:
	move.w	#disable,sr		*** Disable to allocate MCB
	move.l	mcblist,a6		*** a6 -> first message block
	cmpa.l	#0,a6			*** Are there any?
	beq	nomcbs			*** No, return an error
	move.l	(a6),mcblist		*** Remove block from list
	move.w	#enable,sr		*** Allow interrupts
	clr.l	m_link(a6)		*   Clear link word
	move.l	a0,m_buff(a6)		*   Put buffer address into MCB
	move.l	d1,m_count(a6)		*   Put count into MCB
	move.l	current,a1		*   a1 -> Current task
	move.l	a1,m_rtcb(a6)		*   Save task's tcb address
	cmp.l	#maxtask,d2		*   Is d2 in range?
	bhis	badtask			*   No, don't go any further
	bset.b	#MWAIT,t_flag+3(a1)	*   Make task wait for reply
	asl.l	#2,d2			*   multiply task index by 4
	lea	tasktab,a2		*   a2 -> tasktable
	adda.l	d2,a2			*   a2 -> address of TCB address
	move.l	(a2),a2			*   a2 -> Target TCB
	lea	t_msg(a2),a3		*   a3 -> message list
	move.w	#disable,sr		*** Disable for queue search
getmcb:	tst.l	m_link(a3)		*** Null link?
	beq	mcbend			*** Yes, add in at end
	move.l	m_link(a3),a3		*** Advance to next link in list
	bra	getmcb			*** Loop to - getmcb
mcbend:	move.l	a6,m_link(a3)		*** Link MCB in at end
	move.w	#enable,sr		*** Re-allow interrupts
	bclr.b	#RWAIT,t_flag+3(a2)	*   Indicate task dispatchable
	bra	newtask			*   Go to new task.
*
*	Here on error
*
badtask:
	move.l	#e_badtask,(a7)		*   Modify D0 on stack
	bra	dorte			*   Go back to task code
nomcbs:
	move.l	#e_nomcb,(a7)		*** Modify D0 on stack
	bra	dorte			*** Go back to task code

	.data
tasktab:	.dc.l	timetask,termtask,app1task,app2task
maxtask:	.equ	(*-tasktab)/4
i_time:		.equ	0		*   Timer Task index
i_term:		.equ	1		*   Terminal Task index
	.text









************************************************************************
*
*		R e c e i v e  M e s s a g e  S V C
*		===================================
*
* Routine "recv" is used by a task desiring to receive a message from
* another task. Incoming messages are queued up in FIFO order by
* "send". The message is removed from the "t_msg" list and placed
* in the "t_rply" list.
*
************************************************************************

recv:
	move.l	current,a0		*   a0 -> Current task's TCB
	move.w	#disable,sr		*** Disable interrupts
	tst.l	t_msg(a0)		*** Any messages waiting?
	beq	nomsgs			*** No, have to wait for one
	move.l	t_msg(a0),a1		*** a1 -> First message
	move.l	m_link(a1),t_msg(a0)	*** Remove from t_msg list
	move.l	t_rply(a0),m_link(a1)	*** Put on
	move.l	a1,t_rply(a0)		*** 		Reply list
	move.l	a1,(a7)			*** Replace D0 contents
	bra	dorte			*** Go back to task.
nomsgs:	bset.b	#RWAIT,t_flag+3(a0)	*** Mark waiting for message
	move.l	#SVCNEW,d0		*** Issue new task
	trap	#0			*** 			SVC
	bra	recv			*** try receive again
	.page

************************************************************************
*
*		R e p l y  t o  M e s s a g e  S V C
*		====================================
*
* Routine "reply" causes the task which originally sent a message to
* resume execution, with D0 set equal to the returned status code
* from the task which received and processed the message.
*
************************************************************************

reply:
	move.l	current,a2		*   a2 -> current task TCB
	lea	t_rply(a2),a1		*   a1 -> list header
	move.w	#disable,sr		*** No interrupts in search
fndmcb:	tst.l	m_link(a1)		*** End of list?
	beq	badmcb			*** Yes, return error
	cmp.l	m_link(a1),a0		*** link point to right mcb?
	beq	found			*** Yes, take out of list
	move.l	m_link(a1),a1		*** Advance the list
	bra	fndmcb			*** And try again
found:	move.l	m_link(a0),m_link(a1)	*** take out of list
	move.l	m_rtcb(a0),a3		*** a3 -> Sender's TCB
	move.l	t_ssp(a3),a4		*** a4 -> Sender's D0
	move.l	m_stat(a1),(a4)		*** Put in status code
	bclr.b	#MWAIT,t_flag+3(a3)	*** Mark dispatchable
	move.l	mcblist,m_link(a0)	*** Put MCB
	move.l	a0,mcblist		*** 	back in free list
	move.w	#enable,sr		*   Enable interrupts
	bra	newtask			*   And re-dispatch
badmcb:	move.l	#e_badmcb,(a7)		*** Set error code
	bra	dorte			*** return to offender
	.page




************************************************************************
*
*		T e r m i n a l  O u t p u t  S V C  
*		===================================
*
* Routine "prtline" causes a buffer full of output to be sent to the
* terminal. The task is suspended untill the buffer is output.
*
************************************************************************

prtline:
	move.l	#i_term,d2		*   Set appropriate index
	bra	send			*   And do a send message

************************************************************************
*
*			D e l a y  S V C
*			================
*
* Routine "delay" causes the issuing task to be delayed by the number of
* clock ticks contained in register D1.
*
************************************************************************

delay:
	move.l	#i_time,d2		*   Load task index
	bra	send			*   And do a send SVC
	.page






































************************************************************************
*
*			T i m e r  T a s k
*			==================
*
************************************************************************

timeepa:
	move.l	#SVCRCV,d0		*   Receive a message from
	trap	#0			*   user task requesting delay
	move.l	d0,a0			*   a0 -> MCB for request
	move.l	m_count(a0),tcount	*   Copy count for - t_proc
	move.l	current,a1		*   a1 -> our TCB
	move.w	#disable,sr		*** Disable interrupts
	bset.b	#IWAIT,t_flag+3(a1)	*** Mark task non-dispatchable
	bsr	trset			*** Load timer registers
	move.l	#SVCNEW,d0		*** Switch to next
	trap	#0			*** 		task
	move.w	#enable,sr		*   Re-enable interrupts
	move.l	#SVCRPLY,d0		*   Reply to MCB
	trap	#0			*   Issue Reply SVC
	bra	timeepa			*   Go around again
*
* The following code is entered via timer interrupts.
*
timeint:
	move.w	#disable,sr		*** Disable interrupts
	movem.l	d0-d7/a0-a6,-(a7)	*** Save all registers	
	subq.l	#1,tcount		*** Decrement counter
	beq	twake			*** EQ => delay finished
	bra	dorte			*** Otherwise, delay more
twake:
	lea	timetask,a0		*** a0 -> our TCB
	bclr.b	#IWAIT,t_flag+3(a0)	*** Mark task part dispatchable
	bclr.b	#5,IERB			*** disable timer interrupt
	bra	newtask			*** enter dispatcher
*
*	Timer reset routine
*
trset:
	bclr.b	#5,IERB			*** Disable timer interrupt
	move.b	#207,TCDR		*** Put clock in phase
	bset.b	#5,IERB			*** Enable clock interrupts
	rts				*** 

	.bss
tcount:		.ds.l	1		*    Number of ticks remaining
	.text


















************************************************************************
*
*		T e r m i n a l  O u t p u t  T a s k
*		=====================================
*
*	This code performs terminal output for other tasks.
*
************************************************************************

termepa:
	move.l	#SVCRCV,d0		*   Receive output
	trap	#0			*   		request
	move.l	d0,a0			*   a0 -> MCB
	move.l	m_count(a0),bcount	*   Make local copy of
	move.l	m_buff(a0),a2		*   		required data
	subq.l	#1,bcount		*   Decrement count
	move.l	current,a1		*   a1 -> our TCB
	move.w	#disable,sr		*** disable interrupts for start
	bset.b	#IWAIT,t_flag+3(a1)	*** Mark us not dispatchable
	bset.b	#0,IERB			*** Enable terminal interrupt
	bsr	printer			*** Send character to the printer
	move.l	a2,buff			*** Remember buffer address
	move.l	#SVCNEW,d0		*** Execute
	trap	#0			*** 		next task
	move.w	#enable,sr		*   Re_enable interrupts
	move.l	#SVCRPLY,d0		*   Reply to MCB
	trap	#0			*   Issue reply SVC
	bra	termepa			*   Do next buffer
printer:
	move.b	(a2)+,giwrite		*** Output character
	move.b	#$0e,giselect		*** Select PSG port A
	move.b	#$17,giwrite		*** Send strobe low
	move.b	#$17,giwrite		*** for a while
	move.b	#$37,giwrite		*** Send strobe high
	move.b	#$0f,giselect		*** Select PSG port B
	rts				***
*
* The following code is entered by interrupt (centronics busy)
*		for each character outputted.
termint:
	move.w	#disable,sr		*** Disable interrupts
	movem.l	d0-d7/a0-a6,-(a7)	*** Save all registers
	tst.l	bcount			*** Are we done?
	bgt	nextchar		*** No, continue to output
	bclr.b	#0,IERB			*** Disable terminal interrupt
	lea	termtask,a0		*** a0 -> our TCB
	bclr.b	#IWAIT,t_flag+3(a0)	*** Clear wait bit
	bra	newtask			*** Go find a new task
nextchar:
	move.l	buff,a2			*** a2 -> Character to output
	bsr	printer			***		to the printer
	move.l	a2,buff			*** Save next address
	subq.l	#1,bcount		*** decrement count
	bra	dorte			*** Resume interrupted task
*
*	Data area
*
	.bss	
bcount:		.ds.l	1		*   Byte count to output
buff:		.ds.l	1		*   Buffer address to output
	.end
